@@ -14,18 +14,18 @@
 #include "nv-freebsd.h"
 
 static d_open_t  nvidia_ctl_open;
-static void nvidia_ctl_dtor(void *arg);
+static d_close_t nvidia_ctl_close;
 static d_ioctl_t nvidia_ctl_ioctl;
 static d_poll_t  nvidia_ctl_poll;
-static d_mmap_single_t nvidia_ctl_mmap_single;
 
 static struct cdevsw nvidia_ctl_cdevsw = {
     .d_open =      nvidia_ctl_open,
+    .d_close =     nvidia_ctl_close,
     .d_ioctl =     nvidia_ctl_ioctl,
     .d_poll =      nvidia_ctl_poll,
-    .d_mmap_single = nvidia_ctl_mmap_single,
     .d_name =      "nvidiactl",
     .d_version =   D_VERSION,
+    .d_flags =     D_TRACKCLOSE|D_NEEDGIANT
 };
 
 static struct cdev *nvidia_ctl_cdev = NULL;
@@ -33,7 +33,7 @@ struct nvidia_softc nvidia_ctl_sc;
 
 static int nvidia_count = 0;
 
-static int nvidia_ctl_open(
+int nvidia_ctl_open(
     struct cdev *dev,
     int oflags,
     int devtype,
@@ -42,50 +42,32 @@ static int nvidia_ctl_open(
 {
     int status;
     nv_state_t *nv = &nvidia_ctl_state;
-    struct nvidia_filep *filep;
-
-    filep = malloc(sizeof(nvidia_filep_t), M_NVIDIA, (M_WAITOK | M_ZERO));
-    if (filep == NULL)
-        return ENOMEM;
-
-    status = devfs_set_cdevpriv(filep, nvidia_ctl_dtor);
-    if (status != 0) {
-        free(filep, M_NVIDIA);
-        return status;
-    }
-
-    filep->nv = nv;
-    mtx_init(&filep->event_mtx, "event_mtx", NULL, (MTX_DEF | MTX_RECURSE));
-    STAILQ_INIT(&filep->event_queue);
 
     nv_lock_api(nv);
-    status = nvidia_open_ctl(nv, filep);
+    status = nvidia_open_ctl(dev, td);
     nv_unlock_api(nv);
 
     return status;
 }
 
-static void nvidia_ctl_dtor(void *arg)
+int nvidia_ctl_close(
+    struct cdev *dev,
+    int fflag,
+    int devtype,
+    d_thread_t *td
+)
 {
     int status;
-    struct nvidia_filep *filep = arg;
-    struct nvidia_event *et;
-    nv_state_t *nv = filep->nv;
+    nv_state_t *nv = &nvidia_ctl_state;
 
     nv_lock_api(nv);
-    status = nvidia_close_ctl(nv, filep);
+    status = nvidia_close_ctl(dev, td);
     nv_unlock_api(nv);
 
-    while ((et = STAILQ_FIRST(&filep->event_queue))) {
-        STAILQ_REMOVE(&filep->event_queue, et, nvidia_event, queue);
-        free(et, M_NVIDIA);
-    }
-    mtx_destroy(&filep->event_mtx);
-
-    free(filep, M_NVIDIA);
+    return status;
 }
 
-static int nvidia_ctl_ioctl(
+int nvidia_ctl_ioctl(
     struct cdev *dev,
     u_long cmd,
     caddr_t data,
@@ -93,84 +75,105 @@ static int nvidia_ctl_ioctl(
     d_thread_t *td
 )
 {
-    int status;
+    int status = 0;
     nv_state_t *nv = &nvidia_ctl_state;
-    struct nvidia_filep *filep;
-
-    status = devfs_get_cdevpriv((void **)&filep);
-    if (status != 0)
-        return status;
 
     if (__NV_IOC_TYPE(cmd) != NV_IOCTL_MAGIC)
         return ENOTTY;
 
     nv_lock_api(nv);
-    status = nvidia_handle_ioctl(nv, filep, cmd, data);
-    nv_unlock_api(nv);
 
+    switch (__NV_IOC_NR(cmd)) {
+        case NV_ESC_CARD_INFO:
+            do {
+                struct nv_ioctl_card_info *ci = (void *) data;
+                status = nvidia_get_card_info(ci);
+            } while (0);
+            break;
+
+        case NV_ESC_POST_VBIOS:
+            do {
+                struct nv_ioctl_post_vbios *pc = (void *) data;
+                status = nvidia_post_vbios(pc);
+            } while (0);
+            break;
+
+        case NV_ESC_CHECK_VERSION_STR:
+            do {
+                nv_ioctl_rm_api_version_t *params = (void *) data;
+                status = (rm_perform_version_check(NULL, params) == RM_OK) ? 0 : -EINVAL;
+            } while (0);
+            break;
+
+        default:
+            status = nvidia_handle_ioctl(dev, cmd, data, fflag, td);
+    }
+
+    nv_unlock_api(nv);
     return status;
 }
 
-static int nvidia_ctl_poll(
+
+/*
+ * The X driver module adds the ctrl device fd to the XFree86 select list
+ * and receives notification if the select status changes. We make use of
+ * this mechanism to signal hotkey events; this function is called by the
+ * core resource manager whenever a legacy hotkey event is detected.
+ */
+
+void nv_set_hotkey_occurred_flag(void)
+{
+    nv_state_t *nv;
+    struct nvidia_softc *sc;
+
+    nv = &nvidia_ctl_state;
+    sc = nv->os_state;
+
+    nv_lock_rm(nv);
+    nv->flags |= NV_FLAG_HOTKEY_OCCURRED;
+    nv_unlock_rm(nv);
+
+    selwakeup(&sc->rsel);
+}
+
+int nvidia_ctl_poll(
     struct cdev *dev,
     int events,
     d_thread_t *td
 )
 {
-    struct nvidia_filep *filep;
-    int status, mask = 0;
+    nv_state_t *nv;
+    struct nvidia_softc *sc;
 
-    status = devfs_get_cdevpriv((void **)&filep);
-    if (status != 0)
-        return 0;
+    nv = &nvidia_ctl_state;
+    sc = nv->os_state;
 
-    mtx_lock(&filep->event_mtx);
+    nv_lock_rm(nv);
 
-    if (STAILQ_EMPTY(&filep->event_queue) && !filep->event_pending)
-        selrecord(td, &filep->event_rsel);
-    else {
-        mask = (events & (POLLIN | POLLPRI | POLLRDNORM));
-        filep->event_pending = FALSE;
+    if ((nv->flags & NV_FLAG_HOTKEY_OCCURRED) == 0) {
+        nv_unlock_rm(nv);
+        selrecord(td, &sc->rsel);
+    } else {
+        nv->flags &= ~NV_FLAG_HOTKEY_OCCURRED;
+        nv_unlock_rm(nv);
+
+        return (events & (POLLIN | POLLRDNORM)); /* ? */
     }
 
-    mtx_unlock(&filep->event_mtx);
-
-    return mask;
-}
-
-static int nvidia_ctl_mmap_single(
-    struct cdev *dev,
-    vm_ooffset_t *offset,
-    vm_size_t size,
-    vm_object_t *object,
-    int nprot
-)
-{
-    int status;
-    struct nvidia_filep *filep;
-    nv_state_t *nv;
-
-    status = devfs_get_cdevpriv((void **)&filep);
-    if (status != 0)
-        return status;
-    nv = filep->nv;
-
-    nv_lock_api(nv);
-    status = nvidia_mmap_ctl_single(nv, filep, offset, size, object);
-    nv_unlock_api(nv);
-
-    return status;
+    return 0;
 }
 
 int nvidia_ctl_attach(void)
 {
     if (nvidia_count == 0) {
+        /*
+         * This routine is called from nvidia_attach(), multiple times
+         * when more than one device is installed.
+         */
         nvidia_ctl_cdev = make_dev(&nvidia_ctl_cdevsw,
                 CDEV_CTL_MINOR,
                 UID_ROOT, GID_WHEEL, 0666,
                 "%s", nvidia_ctl_cdevsw.d_name);
-        if (nvidia_ctl_cdev == NULL)
-            return ENOMEM;
     }
 
     nvidia_count++;
@@ -184,8 +187,13 @@ int nvidia_ctl_detach(void)
     sc = &nvidia_ctl_sc;
     nvidia_count--;
 
-    if (nvidia_count == 0)
+    if (nvidia_count == 0) {
+        /*
+         * Like nvidia_ctl_attach(), nvidia_ctl_detach() will also be
+         * called more than once with multiple devices.
+         */
         destroy_dev(nvidia_ctl_cdev);
+    }
 
     return 0;
 }
