@@ -13,25 +13,14 @@
 #include "nv.h"
 #include "nv-freebsd.h"
 
-static void nvidia_pci_identify (driver_t *, device_t);
-static int  nvidia_pci_probe    (device_t);
-static int  nvidia_pci_attach   (device_t);
-static int  nvidia_pci_detach   (device_t);
-
-static void nvidia_pci_disable_serr  (device_t dev);
-static void nvidia_pci_enable_intx   (device_t dev);
-
-void nvidia_pci_identify(driver_t *driver, device_t parent)
-{
-    if (device_find_child(parent, "nvidia", -1) == NULL)
-        device_add_child(parent, "nvidia", -1);
-}
+static int nvidia_pci_probe  (device_t);
+static int nvidia_pci_attach (device_t);
+static int nvidia_pci_detach (device_t);
 
 int nvidia_pci_probe(device_t dev)
 {
-    NvU16 device;
+    U016 device;
     char name[NV_DEVICE_NAME_LENGTH];
-    nv_stack_t *sp;
 
     if ((pci_get_class(dev) != PCIC_DISPLAY) ||
             ((pci_get_subclass(dev) != PCIS_DISPLAY_VGA) &&
@@ -40,24 +29,16 @@ int nvidia_pci_probe(device_t dev)
             ((device = pci_get_device(dev)) < 0x0020))
         return ENXIO;
 
-    NV_UMA_ZONE_ALLOC_STACK(sp);
-    if (sp == NULL)
-        return ENOMEM;
-
-    if (rm_is_legacy_device(sp, device, TRUE)) {
-        NV_UMA_ZONE_FREE_STACK(sp);
+    if (rm_is_legacy_device(NULL, device, TRUE)) {
         return ENXIO;
     }
-
-    if (rm_get_device_name(sp, NULL, device, NV_DEVICE_NAME_LENGTH, name)
+    
+    if (rm_get_device_name(NULL, NULL, device, NV_DEVICE_NAME_LENGTH, name)
             != RM_OK) {
         strcpy(name, "Unknown");
     }
 
-    NV_UMA_ZONE_FREE_STACK(sp);
-
     device_set_desc_copy(dev, name);
-
     return 0;
 }
 
@@ -68,11 +49,14 @@ int nvidia_pci_setup_intr(device_t dev)
 
     sc = device_get_softc(dev);
 
-    /* XXX Revisit! (INTR_FAST) */
-    flags = (INTR_TYPE_AV | INTR_MPSAFE);
+    /* XXX Revisit! (INTR_FAST, INTR_MPSAFE) */
+    flags = INTR_TYPE_AV;
 
-    status = bus_setup_intr(dev, sc->irq, flags, NULL, nvidia_intr, sc,
-            &sc->irq_ih);
+#if __FreeBSD_version >= 700031
+    status = bus_setup_intr(dev, sc->irq, flags, NULL, nvidia_intr, sc, &sc->irq_ih);
+#else
+    status = bus_setup_intr(dev, sc->irq, flags, nvidia_intr, sc, &sc->irq_ih);
+#endif
     if (status) {
         device_printf(dev, "NVRM: HW ISR setup failed.\n");
         goto fail;
@@ -99,36 +83,15 @@ fail:
     return status;
 }
 
-static void nvidia_pci_disable_serr(device_t dev)
+void nvidia_pci_check_config_space(device_t dev, BOOL do_the_bars)
 {
-    NvU16 word;
-    word = (os_pci_read_word(dev, PCIR_COMMAND) & ~PCIM_CMD_SERRESPEN);
-    os_pci_write_word(dev, PCIR_COMMAND, word);
-}
-
-static void nvidia_pci_enable_intx(device_t dev)
-{
-    NvU16 word;
-    word = (os_pci_read_word(dev, PCIR_COMMAND) & ~PCIM_CMD_INTXDIS);
-    os_pci_write_word(dev, PCIR_COMMAND, word);
-}
-
-void nvidia_pci_check_config_space(
-    nv_stack_t *sp,
-    device_t dev,
-    BOOL check_the_bars,
-    BOOL acquire_semaphore,
-    BOOL may_block
-)
-{
-    NvU32 BAR_low, BAR_high;
-    NvU16 word, i;
+    U032 BAR_low, BAR_high;
+    U016 word, i;
     struct nvidia_softc *sc = device_get_softc(dev);
     nv_state_t *nv = sc->nv_state;
 
     if (nv->flags & NV_FLAG_USE_BAR0_CFG) {
-        rm_check_pci_config_space(sp, nv, check_the_bars,
-            acquire_semaphore, may_block);
+        rm_check_pci_config_space(NULL, nv, do_the_bars, NULL);
         return;
     }
 
@@ -140,13 +103,7 @@ void nvidia_pci_check_config_space(
     if ((word & PCIM_CMD_MEMEN) == 0)
         pci_enable_io(dev, SYS_RES_MEMORY);
 
-    if ((word & PCIM_CMD_SERRESPEN) != 0)
-        nvidia_pci_disable_serr(dev);
-
-    if ((word & PCIM_CMD_INTXDIS) != 0)
-        nvidia_pci_enable_intx(dev);
-
-    if (check_the_bars)
+    if (do_the_bars)
     {
         for (i = 0; i < NV_GPU_NUM_BARS; i++) {
             BAR_low = BAR_high = 0;
@@ -172,10 +129,10 @@ void nvidia_pci_check_config_space(
     }
 }
 
-NvU8 nvidia_pci_find_capability(device_t dev, NvU8 capability)
+U008 nvidia_pci_find_capability(device_t dev, U008 capability)
 {
-    NvU16 status;
-    NvU8 cap_ptr, cap_id;
+    U016 status;
+    U008 cap_ptr, cap_id;
 
     status = pci_read_config(dev, PCIR_STATUS, 2);
     status &= PCIM_STATUS_CAPPRESENT;
@@ -208,9 +165,8 @@ int nvidia_pci_attach(device_t dev)
 {
     int status;
     struct nvidia_softc *sc;
-    NvU16 word, i, j;
-    NvU32 BAR_low, req;
-    nv_stack_t *sp;
+    U016 word, i, j;
+    U032 BAR_low, req;
 
     if (device_get_unit(dev) >= NV_MAX_DEVICES) {
         device_printf(dev, "NVRM: maximum device number exceeded.\n");
@@ -218,7 +174,9 @@ int nvidia_pci_attach(device_t dev)
     }
 
     sc = device_get_softc(dev); /* first reference */
+
     bzero(sc, sizeof(nvidia_softc_t));
+    STAILQ_INIT(&sc->filep_queue);
 
     sc->nv_state = malloc(sizeof(nv_state_t), M_NVIDIA, M_WAITOK | M_ZERO);
     if (sc->nv_state == NULL)
@@ -241,7 +199,7 @@ int nvidia_pci_attach(device_t dev)
     }
 
     for (i = 0, j = 0; i < NVRM_PCICFG_NUM_BARS && j < NV_GPU_NUM_BARS; i++) {
-        NvU8 offset = NVRM_PCICFG_BAR_OFFSET(i);
+        U008 offset = NVRM_PCICFG_BAR_OFFSET(i);
         BAR_low = os_pci_read_dword(dev, offset);
         os_pci_write_dword(dev, offset, 0xffffffff);
         req = os_pci_read_dword(dev, offset);
@@ -258,14 +216,6 @@ int nvidia_pci_attach(device_t dev)
     sc->irq_rid = 0;
     sc->iop_rid = 0;
 
-    NV_UMA_ZONE_ALLOC_STACK(sp);
-    if (sp == NULL) {
-        free(sc->nv_state, M_NVIDIA);
-        return ENOMEM;
-    }
-
-    sc->attach_sp = sp;
-
     status = nvidia_alloc_hardware(dev);
     if (status) {
         device_printf(dev, "NVRM: NVIDIA hardware alloc failed.\n");
@@ -285,33 +235,27 @@ int nvidia_pci_attach(device_t dev)
         goto fail;
     }
 
-    if (!rm_init_private_state(sp, sc->nv_state)) {
+    if (!rm_init_private_state(NULL, sc->nv_state)) {
         nvidia_pci_teardown_intr(dev);
         device_printf(dev, "NVRM: rm_init_private_state() failed.\n");
         nvidia_detach(dev);
         goto fail;
     }
 
-    callout_init(&sc->timer, CALLOUT_MPSAFE);
+    mtx_init(&sc->rm_mtx, "dev.rm_mtx", NULL, MTX_SPIN | MTX_RECURSE);
     sx_init(&sc->api_sx, "dev.api_sx");
 
     return 0;
 
 fail:
     nvidia_free_hardware(dev);
-
-    sc->attach_sp = NULL;
-    NV_UMA_ZONE_FREE_STACK(sp);
-
     free(sc->nv_state, M_NVIDIA);
-
     return status;
 }
 
 int nvidia_pci_detach(device_t dev)
 {
     int status;
-    nv_stack_t *sp;
     struct nvidia_softc *sc;
     nv_state_t *nv;
 
@@ -331,6 +275,8 @@ int nvidia_pci_detach(device_t dev)
     }
 
     nv_unlock_api(nv);
+
+    mtx_destroy(&sc->rm_mtx);
     sx_destroy(&sc->api_sx);
 
     status = nvidia_pci_teardown_intr(dev);
@@ -343,14 +289,8 @@ int nvidia_pci_detach(device_t dev)
         goto fail;
     }
 
-    sp = sc->attach_sp;
-
-    rm_free_private_state(sp, sc->nv_state);
+    rm_free_private_state(NULL, sc->nv_state);
     nvidia_free_hardware(dev);
-
-    sc->attach_sp = NULL;
-    NV_UMA_ZONE_FREE_STACK(sp);
-
     free(sc->nv_state, M_NVIDIA);
 
 fail:
@@ -359,13 +299,12 @@ fail:
 }
 
 static device_method_t nvidia_pci_methods[] = {
-    DEVMETHOD( device_identify, nvidia_pci_identify ),
-    DEVMETHOD( device_probe,    nvidia_pci_probe    ),
-    DEVMETHOD( device_attach,   nvidia_pci_attach   ),
-    DEVMETHOD( device_detach,   nvidia_pci_detach   ),
+    DEVMETHOD( device_probe,   nvidia_pci_probe  ),
+    DEVMETHOD( device_attach,  nvidia_pci_attach ),
+    DEVMETHOD( device_detach,  nvidia_pci_detach ),
 #ifdef NV_SUPPORT_ACPI_PM
-    DEVMETHOD( device_suspend,  nvidia_suspend      ),
-    DEVMETHOD( device_resume,   nvidia_resume       ),
+    DEVMETHOD( device_suspend, nvidia_suspend    ),
+    DEVMETHOD( device_resume,  nvidia_resume     ),
 #endif
     { 0, 0 }
 };
@@ -376,7 +315,7 @@ static driver_t nvidia_pci_driver = {
     sizeof(struct nvidia_softc)
 };
 
-DRIVER_MODULE(nvidia, vgapci, nvidia_pci_driver, nvidia_devclass, nvidia_modevent, 0);
+DRIVER_MODULE(nvidia, pci, nvidia_pci_driver, nvidia_devclass, nvidia_modevent, 0);
 
 MODULE_DEPEND(nvidia, mem, 1, 1, 1);
 MODULE_DEPEND(nvidia, io, 1, 1, 1);
@@ -388,3 +327,4 @@ MODULE_DEPEND(nvidia, agp, 1, 1, 1);
 #ifdef NV_SUPPORT_LINUX_COMPAT /* (COMPAT_LINUX || COMPAT_LINUX32) */
 MODULE_DEPEND(nvidia, linux, 1, 1, 1);
 #endif
+
